#include "consts.h"
#include "dirichlet.h"
#include "error.h"
#include "files.h"
#include "io_contnr.h"
#include "partition.h"

#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <deque>
#include <map>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_multimin.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf_gamma.h>
#include <boost/multi_array.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/program_options.hpp>
#include <boost/assign/list_inserter.hpp>
#include <boost/assign/list_of.hpp>
#include <boost/assign/std/vector.hpp>
#include <boost/assign/std/set.hpp>
#include <boost/foreach.hpp>
#include <boost/format.hpp>
#include <boost/regex.hpp>	
#include <arprec/mp_real.h>

using namespace std;
using namespace boost;
using namespace boost::program_options;
using namespace boost::assign;
using namespace partition;

#define FOREACH BOOST_FOREACH
#define IS_NAN( x ) ((x) != (x))
#define COMMA ,

ofstream output(output_file_name().c_str());

struct Seq { string genome; int no; string sub_name; };

map<string, int> sub_name_to_idx;
map<int, string> idx_to_sub_name;
vector<Seq> seqs;
map<char, int> nucl_to_idx;



// read alignment from file
void get_alignment() 
{
	ifstream AL_FILE(alignment_file_name().c_str());
	if (!AL_FILE) {
		string str = "Could not open " + alignment_file_name();
		throw Error(str);
	}	
	
	set<string> sub_names;

	string str;
	string err_str = "File format error in " + alignment_file_name();
	bool is_sub = true;
	string curr_sub_name;
	int curr_no = 1;
	bool is_finished = false;
	int idx = 0;
	while(!is_finished) {
		if (AL_FILE.eof())
			is_finished = true;
		else {			
			getline(AL_FILE, str);
			if (str.length()) {
				if (str[0] == '>') {
					if (!is_sub)				
						throw Error(err_str);			
					curr_sub_name = str.substr(1);			
				}
				else {
					if (is_sub)				
						throw Error(err_str);	
					if (sub_names.find(curr_sub_name) == sub_names.end()) {
						sub_name_to_idx[curr_sub_name] = idx;
						idx_to_sub_name[idx] = curr_sub_name;
						++idx;
					}						
					sub_names += curr_sub_name;
					Seq seq = { str, curr_no, curr_sub_name };
					seqs += seq;
					++curr_no;
				}
				is_sub = !is_sub;
			}
		}
	}		
}



// read begin and end position and small and large subtypes and numbers of seqs to be used
void get_base_data(unsigned* begin_pos, unsigned* end_pos, unsigned* seq_len, set<unsigned>* small_subs, set<unsigned>* large_subs, set<unsigned>* no_used_seqs)
{
	ifstream SEQS_FILE(seqs_file_name().c_str());
	if (!SEQS_FILE) {
		string str = "Could not open " + seqs_file_name();
		throw Error(str);
	}		
	
	string str;
	getline(SEQS_FILE, str);
	stringstream ss(str);
	ss >> *begin_pos;

	getline(SEQS_FILE, str);
	stringstream tt(str);
	tt >> *end_pos;
	*seq_len = *end_pos - *begin_pos + 1;
	
	getline(SEQS_FILE, str);
	if (str != ">small") {
		string str = "File format error (>small) in " + seqs_file_name();
		throw Error(str);
	}
	
	bool is_finished = false;
	while (!is_finished) {
		getline(SEQS_FILE, str);
		if (str[0] == '>')
			is_finished = true;
		else {
			*small_subs += sub_name_to_idx[str]; 
		}
	}
	if (str != ">large") {
		string str = "File format error (>large) in " + seqs_file_name();
		throw Error(str);
	}	
	
	is_finished = false;
	while (!is_finished) {
		getline(SEQS_FILE, str);
		if (str[0] == '>')
			is_finished = true;
		else {
			*large_subs += sub_name_to_idx[str]; 
		}
	}
	if (str != ">no_seqs") {
		string str = "File format error (>no_seqs) in " + seqs_file_name();
		throw Error(str);
	}		
	
	is_finished = false;
	while (!is_finished) {
		getline(SEQS_FILE, str);
		if (SEQS_FILE.eof())
			is_finished = true;
		else {		
			int i;
			stringstream ss(str);
			ss >> i;
			*no_used_seqs += i;
		}
	}	
}



// calculate emission freqs 
void calc_em_freq(mul_arr_int_3* em_freq, unsigned* seq_len, set<unsigned>* small_subs, set<unsigned>* large_subs)
{
	const bool OUTPUT_DEBUG = 0;
	const bool OUTPUT_TO_FILE = 1;

	unsigned num_small_subs, num_large_subs, num_subs;

	unsigned begin_pos, end_pos;
	set<unsigned> no_used_seqs;
	get_base_data(&begin_pos, &end_pos, seq_len, small_subs, large_subs, &no_used_seqs);
	
	num_small_subs = small_subs->size();
	num_large_subs = large_subs->size();
	num_subs = num_small_subs + num_large_subs;
	mul_arr_int_3::extent_gen extents;
	em_freq->resize(extents[*seq_len][num_subs][Consts::num_nucl()]);
	for (unsigned j = 0; j < num_subs; ++j)
		for (unsigned i = 0; i < *seq_len; ++i)
			for (unsigned k = 0; k < Consts::num_nucl(); ++k)
				(*em_freq)[i][j][k] = 0;

	FOREACH (Seq seq, seqs) {
		if (no_used_seqs.find(seq.no) != no_used_seqs.end()) {
			int sub_idx = sub_name_to_idx[seq.sub_name];
			for (unsigned i = 0; i < *seq_len; ++i) {
				map<char, int>::iterator it = nucl_to_idx.find(seq.genome[begin_pos + i]);
				if (it != nucl_to_idx.end()) {
					int nucl_idx = it->second;
					++(*em_freq)[i][sub_idx][nucl_idx];
				}
			}		
		}
	}

	if (OUTPUT_DEBUG) {
		for (unsigned j = 0; j < num_subs; ++j)
			for (unsigned i = 0; i < *seq_len; ++i) {
				cout << "em_freq[" << i << "][" << j << "][0-" << Consts::num_nucl() << "]";
				for (unsigned k = 0; k < Consts::num_nucl(); ++k)
					cout << " " << (*em_freq)[i][j][k];
				cout << endl;
			}
		cout << "seq_len: " << *seq_len << endl;
		cout << "small_subs: ";
		FOREACH (int i, *small_subs)
			cout << i << " ";
		cout << endl;
		cout << "large_subs: ";
		FOREACH (int i, *large_subs)
			cout << i << " ";
		cout << endl;
	}

	if (OUTPUT_TO_FILE) {
		ofstream EF_FILE(EF_file_name().c_str());
		for (unsigned j = 0; j < num_subs; ++j)
			for (unsigned i = 0; i < *seq_len; ++i) {
				EF_FILE << i << " " << j;
				for (unsigned k = 0; k < Consts::num_nucl(); ++k)
					EF_FILE << " " << (*em_freq)[i][j][k];
				EF_FILE << endl;
			}
	}
}



// calculate representative emission frequencies (for different subtypes)
void calc_repr_em_freq(mul_arr_int_3* repr_em_freq, unsigned* seq_len, unsigned* num_subs)
{
	const bool OUTPUT_DEBUG = 0;
	
	*seq_len = seqs.front().genome.size();
	*num_subs = sub_name_to_idx.size();
	
	mul_arr_int_3::extent_gen extents;
	repr_em_freq->resize(extents[*seq_len][*num_subs][Consts::num_nucl()]);
	for (unsigned j = 0; j < *num_subs; ++j)
		for (unsigned i = 0; i < *seq_len; ++i)
			for (unsigned k = 0; k < Consts::num_nucl(); ++k)
				(*repr_em_freq)[i][j][k] = 0;

	FOREACH (Seq seq, seqs) {
		unsigned sub_idx = sub_name_to_idx[seq.sub_name];
		for (unsigned i = 0; i < *seq_len; ++i) {
			map<char, int>::iterator it = nucl_to_idx.find(seq.genome[i]);
			if (it != nucl_to_idx.end()) {
				int nucl_idx = it->second;
				++(*repr_em_freq)[i][sub_idx][nucl_idx];
			}
		}		
	}

	if (OUTPUT_DEBUG) {
		for (unsigned j = 0; j < *num_subs; ++j)
			for (unsigned i = 0; i < *seq_len; ++i) {
				cout << "repr_em_freq[" << i << "][" << j << "][0-" << Consts::num_nucl() << "]";
				for (unsigned k = 0; k < Consts::num_nucl(); ++k)
					cout << " " << (*repr_em_freq)[i][j][k];
				cout << endl;
			}
	}
}



// read representative emission frequencies from file
void get_repr_em_freq(mul_arr_int_2* repr_em_freq, unsigned* num_pos_repr)
{
	const bool OUTPUT_DEBUG = 0;
	const unsigned int INPUT_SIZE = 2 + Consts::num_nucl();

	ifstream REPR_EF_FILE(repr_EF_file_name().c_str());
	if (!REPR_EF_FILE)
		throw Error(string("Could not open ") + repr_EF_file_name());
	if (!REPR_EF_FILE.eof()) {
		string str;
		getline(REPR_EF_FILE, str);
		stringstream ss(str);
		ss >> *num_pos_repr;
	}
	else
		throw Error("File format error in get_repr_em_freq");

	mul_arr_int_2::extent_gen extents;
	repr_em_freq->resize(extents[*num_pos_repr][Consts::num_nucl()]);
	
	int n = 0;
	while (!REPR_EF_FILE.eof()) {
		string str;
		vector<int> input;
		vector<string> part_str;
		getline(REPR_EF_FILE, str);
		if (str.size() > 0) {
			split(part_str, str, is_any_of(" \n") );
			for (vector<string>::iterator it = part_str.begin(); it != part_str.end(); ++it)
				input.push_back(atoi(it->c_str()));
			if (input.size() != INPUT_SIZE)
				throw Error("Error in get_repr_em_freq");
			for (unsigned i = 0; i < Consts::num_nucl(); ++i)
				(*repr_em_freq)[n][i] = input[i + 1];
		}
		++n;
	}

	if (OUTPUT_DEBUG)
		for (unsigned i = 0; i < *num_pos_repr; ++i)
			for (unsigned j = 0; j < Consts::num_nucl(); ++j)
					cout << "repr_em_freq[" << i << "][" << j << "]: " << (*repr_em_freq)[i][j] << endl;
}



// read representative transition frequencies from file
void get_repr_tr_freq(mul_arr_int_2* repr_tr_freq, int* num_pos_repr)
{
	const bool OUTPUT_DEBUG = 0;
	const int NUM_FREQS = 2;
	const unsigned int INPUT_SIZE = 2 + NUM_FREQS;

	ifstream REPR_TF_FILE(repr_TF_file_name().c_str());
	if (!REPR_TF_FILE)
		throw Error(string("Could not open ") + repr_TF_file_name());
	if (!REPR_TF_FILE.eof()) {
		string str;
		getline(REPR_TF_FILE, str);
		stringstream ss(str);
		ss >> *num_pos_repr;
	}
	else
		throw Error("File format error in get_repr_tr_freq");

	mul_arr_int_2::extent_gen extents;
	repr_tr_freq->resize(extents[*num_pos_repr][NUM_FREQS]);

	int n = 0;
	while (!REPR_TF_FILE.eof()) {
		string str;
		vector<int> input;
		vector<string> part_str;
		getline(REPR_TF_FILE, str);
		if (str.size() > 0) {
			split(part_str, str, is_any_of(" \n") );
			for (vector<string>::iterator it = part_str.begin(); it != part_str.end(); ++it)
				input.push_back(atoi(it->c_str()));
			if (input.size() != INPUT_SIZE)
				throw Error("Error in get_repr_tr_freq");
			for (int i = 0; i < NUM_FREQS; ++i)
				(*repr_tr_freq)[n][i] = input[i + 1];
		}
		++n;
	}

	if (OUTPUT_DEBUG)
		for (int i = 0; i < *num_pos_repr; ++i)
			for (int j = 0; j < 2; ++j)
					cout << "repr_tr_freq[" << i << "][" << j << "]: " << (*repr_tr_freq)[i][j] << endl;
}



// calculate transition frequencies
void calc_tr_freq(mul_arr_int_3* tr_freq, unsigned* seq_len, set<unsigned>* small_subs, set<unsigned>* large_subs)
{
	const bool OUTPUT_DEBUG = 0;

	unsigned begin_pos, end_pos;
	set<unsigned> no_used_seqs;
	get_base_data(&begin_pos, &end_pos, seq_len, small_subs, large_subs, &no_used_seqs);	
	unsigned num_subs = small_subs->size() + large_subs->size();
	
	mul_arr_int_3::extent_gen extents;
	tr_freq->resize(extents[*seq_len - 1][num_subs][Consts::num_tr_freqs()]);	
	
	for (unsigned i = 0; i < *seq_len - 1; ++i)
		for (unsigned j = 0; j < num_subs; ++j)
			for (unsigned k = 0; k < Consts::num_tr_freqs(); ++k)
				(*tr_freq)[i][j][k] = 0;
	
	vector<bool> is_cons_col(*seq_len, false);
	for (unsigned i = 0; i < *seq_len; ++i) {
		for (unsigned j = 0; j < seqs.size(); ++j)
			if (nucl_to_idx.find(seqs[j].genome[i]) != nucl_to_idx.end()) {
				is_cons_col[i] = true;
				break;
			}
	}		
	
	unsigned i = begin_pos;
	while (!is_cons_col[i] && (i < end_pos))
		++i;
	
	while (i < end_pos) {	
		unsigned j = i + 1;
        while (!is_cons_col[j - begin_pos] && (j <= end_pos))
            ++j;
        if (j <= end_pos) {
            FOREACH (Seq seq, seqs) {
        		if (no_used_seqs.find(seq.no) != no_used_seqs.end()) {
        			int sub_idx = sub_name_to_idx[seq.sub_name];
            	        			
	                bool source_is_delete = (nucl_to_idx.find(seq.genome[i]) == nucl_to_idx.end()); 	                
	                bool dest_is_delete = (nucl_to_idx.find(seq.genome[j]) == nucl_to_idx.end()); 

	                if (source_is_delete) {
	                    if (dest_is_delete)
	                        ++(*tr_freq)[i - begin_pos][sub_idx][3];
	                    else
	                    	++(*tr_freq)[i - begin_pos][sub_idx][2];
	                }
	                else {
	                    if (dest_is_delete)
	                    	++(*tr_freq)[i - begin_pos][sub_idx][1];
	                    else
	                    	++(*tr_freq)[i - begin_pos][sub_idx][0];
	            	}
	        	}
        	}
        }
        i = j;
	}
	
	if (OUTPUT_DEBUG) {
		for (unsigned i = 0; i < *seq_len - 1; ++i)
			for (unsigned j = 0; j < num_subs; ++j)
				for (unsigned k = 0; k < Consts::num_tr_freqs(); ++k)
					cout << "tr_freq[" << i << "][" << j << "][" << k << "]: " << (*tr_freq)[i][j][k] << endl;
		cout << "seq_len: " << *seq_len << endl;
		cout << "small_subs: ";
		FOREACH (int i, *small_subs)
			cout << i << " ";
		cout << endl;
		cout << "large_subs: ";
		FOREACH (int i, *large_subs)
			cout << i << " ";
		cout << endl;		
	}
}



// get parameter for estimation of emission probabilities
void get_para(map<pair<int, int>, double>* para)
{
	const bool OUTPUT_DEBUG = 0;
	const unsigned int INPUT_SIZE = 3;

	ifstream PARA_FILE(para_file_name().c_str());
	if (!PARA_FILE) {
		string str = "Could not open " + para_file_name();
		throw Error(str);
	}
	
	while (!PARA_FILE.eof()) {
		string str;
		vector<double> input;
		vector<string> part_str;
		getline(PARA_FILE, str);
		if (str.size() > 0) {
			split(part_str, str, is_any_of(" \n") );
			for (vector<string>::iterator it = part_str.begin(); it != part_str.end(); ++it)
				input.push_back(atof(it->c_str()));

			if (input.size() != INPUT_SIZE)
				throw Error("Error in get_para");			
			(*para)[pair<int, int>((int)input[0], (int)input[1])] = input[2];
			if (OUTPUT_DEBUG) {
				cout << "para[" << input[0] << ", " << input[1] << "] = " << input[2] << endl;
			}
		}
	}
}



// save parameter estimates for emission probabilities of representative data
void save_repr_em_para(unsigned num_dir_mix, const mul_arr_dbl_2 alpha, vector<double> q)
{
	ofstream file_strm(repr_em_para_file_name().c_str());
	file_strm << num_dir_mix << endl;

	for (unsigned i = 0; i < num_dir_mix; ++i) {
		for (unsigned j = 0; j < Consts::num_nucl(); ++j) {
			file_strm << alpha[i][j];
			if (j < Consts::num_nucl() - 1)
				file_strm << " ";
		}
		file_strm << endl;
	}

	for (unsigned i = 0; i < num_dir_mix; ++i) {
		file_strm << q[i];
		if (i < num_dir_mix - 1)
			file_strm << " ";
	}
	file_strm << endl;
}



// load parameter estimates for emission probabilities of representative data
void load_repr_em_para(unsigned* num_dir_mix, mul_arr_dbl_2* alpha, vector<double>* q)
{
	ifstream file_strm(repr_em_para_file_name().c_str());
	if (!file_strm) {
		string str = "Could not open " + repr_em_para_file_name();
		throw Error(str);
	}

	file_strm >> *num_dir_mix;

	for (unsigned i = 0; i < *num_dir_mix; ++i)
		for (unsigned j = 0; j < Consts::num_nucl(); ++j)
			file_strm >> (*alpha)[i][j];

	for (unsigned i = 0; i < *num_dir_mix; ++i)
		file_strm >> (*q)[i];
}



// save parameter estimates for transition probabilities of representative data
void save_repr_tr_para(const int num_dir_mix, const mul_arr_dbl_2 alpha, const vector<double> q)
{
	const int NUM_FREQS = 2;
	ofstream file_strm(repr_tr_para_file_name().c_str());
	file_strm << num_dir_mix << endl;

	for (int i = 0; i < num_dir_mix; ++i) {
		for (int j = 0; j < NUM_FREQS; ++j) {
			file_strm << alpha[i][j];
			if (j < NUM_FREQS - 1)
				file_strm << " ";
		}
		file_strm << endl;
	}

	for (int i = 0; i < num_dir_mix; ++i) {
		file_strm << q[i];
		if (i < num_dir_mix - 1)
			file_strm << " ";
	}
	file_strm << endl;
}



// load parameter estimates for transition probabilities of representative data
void load_repr_tr_para(unsigned* num_dir_mix, mul_arr_dbl_2* alpha, vector<double>* q)
{
	const unsigned NUM_FREQS = 2;
	ifstream file_strm(repr_tr_para_file_name().c_str());
	if (!file_strm) {
		string str = "Could not open " + repr_tr_para_file_name();
		throw Error(str);
	}
	file_strm >> *num_dir_mix;

	for (unsigned i = 0; i < *num_dir_mix; ++i)
		for (unsigned j = 0; j < NUM_FREQS; ++j)
			file_strm >> (*alpha)[i][j];

	for (unsigned i = 0; i < *num_dir_mix; ++i)
		file_strm >> (*q)[i];
}



// calculate emission probabilities for strategy 'prob'
void calc_ep_prob(const mul_arr_int_3 em_freq, unsigned seq_len, set<unsigned> small_subs, set<unsigned> large_subs)
{
	const bool OUTPUT_DEBUG = 0;
	const bool OUTPUT = 1;
	
	set<unsigned> subs_set(small_subs.begin(), small_subs.end());
	subs_set.insert(large_subs.begin(), large_subs.end());
	vector<unsigned> subs;
	copy(subs_set.begin(), subs_set.end(), back_inserter(subs));
	mul_arr_dbl_3 p_em;
	mul_arr_dbl_3::extent_gen extents;
	p_em.resize(extents[subs.size()][seq_len][Consts::num_nucl()]);		
	
	double Z = 0;
	double sum_alpha = 0;
	for (unsigned i = 0; i < Consts::num_nucl(); ++i) {
		Z += gsl_sf_lngamma(Consts::alpha()[i]);
		sum_alpha += Consts::alpha()[i];
	}
	Z -= gsl_sf_lngamma(sum_alpha);
	Z = exp(Z);
	if (OUTPUT_DEBUG)
		cout << "Z: " << Z << endl << endl;

	unsigned max_num_distr = min(Consts::max_num_distr(), (unsigned)subs.size());
	for (unsigned pos = 0; pos < seq_len; ++pos) {
		if (OUTPUT)
			cout << "pos: " << pos << endl;

		mul_arr_int_2 N;
		mul_arr_int_2::extent_gen extents;
		N.resize(extents[subs.size()][Consts::num_nucl()]);
		for (unsigned k = 0; k < subs.size(); ++k) {
			for (unsigned m = 0; m < Consts::num_nucl(); ++m) {								
				N[k][m] = em_freq[pos][k][m];									
			}
		}
		
		if (OUTPUT_DEBUG)
			for (unsigned k = 0; k < subs.size(); ++k)
				for (unsigned m = 0; m < Consts::num_nucl(); ++m)
					cout << "N[" << k << "][" << m <<"] = " << N[k][m] << endl;					

		vector<double> a(subs.size());				
		for (unsigned i = 0; i < subs.size(); ++i) {
			double r = 0;
			unsigned N_sub = 0;
			for (unsigned m = 0; m < Consts::num_nucl(); ++m) {		
				N_sub += N[i][m];
				r -= gsl_sf_lngamma(N[i][m] + 1);	
			}
			r += gsl_sf_lngamma(N_sub + 1);
			a[i] = r;
		}
		
		if (OUTPUT_DEBUG)
			cout << "a: " << a << endl;		
		
		vector< map<unsigned, set<unsigned> > > source_combs = get_all_source_partitions(2, subs.size());
		source_combs += get_all_source_partitions(1, subs.size()).front();
		map<set<unsigned>, double> b, c, c_plus;
		vector< map<set<unsigned>, double > > b_plus(Consts::num_nucl());
		
		for (vector< map<unsigned, set<unsigned> > >::iterator it = source_combs.begin(); it != source_combs.end(); ++it) {
			map<unsigned, set<unsigned> > source_comb = *it;
			//cout << "source_comb: " << source_comb << endl; 
			for (map<unsigned, set<unsigned> >::iterator jt = source_comb.begin(); jt != source_comb.end(); ++jt) {
				pair<unsigned, set<unsigned> > source = *jt;
				//cout << "source: " << source.second << endl; 			
				unsigned N_total = 0;
				double x = 0;						
				for (unsigned i = 0; i < Consts::num_nucl(); ++i) {
					unsigned N_nucl = 0;
					FOREACH (unsigned s, source.second)
						N_nucl += N[s][i];
					x += gsl_sf_lngamma(N_nucl + Consts::alpha()[i]);
					N_total += N_nucl;
				}
				b[source.second] = x;
				
				for (unsigned nucl = 0; nucl < Consts::num_nucl(); ++nucl) {
					double x = 0;						
					for (unsigned i = 0; i < Consts::num_nucl(); ++i) {
						unsigned N_nucl = (i == nucl);
						FOREACH (unsigned s, source.second)
							N_nucl += N[s][i];
						x += gsl_sf_lngamma(N_nucl + Consts::alpha()[i]);
					}
					b_plus[nucl][source.second] = x;
				}
				c[source.second] = gsl_sf_lngamma(N_total + sum_alpha);
				c_plus[source.second] = gsl_sf_lngamma(N_total + sum_alpha + 1);
			}
		}
		
		if (OUTPUT_DEBUG) {
			cout << "b: " << endl;
			FOREACH (pair<set<unsigned> COMMA double> pr, b) {
				cout << pr.first << " => " << pr.second << " (" << exp(pr.second) << ")" << endl;
			}
			cout << endl;
			
			for (unsigned m = 0; m < Consts::num_nucl(); ++m) {
				cout << "b_plus(" << m << "): " << endl;
				FOREACH (pair<set<unsigned> COMMA double> pr, b_plus[m])
					cout << pr.first << " => " << pr.second << " (" << exp(pr.second) << ")" << endl;
				cout << endl;
			}
			
			cout << "c: " << endl;
			FOREACH (pair<set<unsigned> COMMA double> pr, c) {
				cout << pr.first << " => " << pr.second << " (" << exp(pr.second) << ")" << endl;
			}
			cout << endl;
			
			cout << "c_plus: " << endl;
			FOREACH (pair<set<unsigned> COMMA double> pr, c_plus) {
				cout << pr.first << " => " << pr.second << " (" << exp(pr.second) << ")" << endl;
			}
			cout << endl;
		}
		
		double prod_a = 0;
		for (unsigned sub = 0; sub < subs.size(); ++sub)
			prod_a += a[sub];

		if (OUTPUT_DEBUG)
			cout << "prod_a: " << prod_a << " (" << exp(prod_a) << ")" << endl << endl << endl;		
		
		mul_arr_dbl_2 p(extents[subs.size()][Consts::num_nucl()]);
		
		for (unsigned num_distr = 1; num_distr <= max_num_distr; ++num_distr) {
			if (OUTPUT_DEBUG)
				cout << "num_distr: " << num_distr << endl;
			vector< map<unsigned, set<unsigned> > > source_combs = get_all_source_partitions(num_distr, subs.size());
			if (OUTPUT_DEBUG) 
				cout << "source_combs: " << source_combs << endl;			
			double prod_Z = log(pow(Z, (double)num_distr));
			double kappa = prod_a + log(Consts::rho()[num_distr - 1]) - prod_Z - log(source_combs.size());
			if (OUTPUT_DEBUG) {
				cout << "prod_Z: " << prod_Z << " (" << exp(prod_Z) << ")" << endl;
				cout << "kappa: " << kappa << " (" << exp(kappa) << ")" << endl;
			}
			
			for (vector< map<unsigned, set<unsigned> > >::iterator it = source_combs.begin(); it != source_combs.end(); ++it) {
				map<unsigned, set<unsigned> > source_comb = *it;
				//cout << "source_comb: " << source_comb << endl; 
				for (map<unsigned, set<unsigned> >::iterator jt = source_comb.begin(); jt != source_comb.end(); ++jt) {
					pair<unsigned, set<unsigned> > source = *jt;
					//cout << "source: " << source.second << endl; 
					unsigned act_source = source.first;	
					double q = 0;
					FOREACH (pair<unsigned COMMA set<unsigned> > pr, source_comb) {
						if (pr.first != act_source)
							q += b[pr.second] - c[pr.second];
					}
					//cout << "q: " << exp(q) << endl; 
					vector<double> p_src(Consts::num_nucl());
					for (unsigned nucl = 0; nucl < Consts::num_nucl(); ++nucl) {
						p_src[nucl] = q + b_plus[nucl][source.second] - c_plus[source.second];
						//cout << "p_src[" << nucl << "]: " << exp(p_src[nucl]) << " (" << b_plus[nucl][source.second] << " " << c_plus[source.second] << ")" << endl; 
					}
					FOREACH (unsigned sub, source.second)
						for (unsigned nucl = 0; nucl < Consts::num_nucl(); ++nucl) {
							p[sub][nucl] += exp(kappa + p_src[nucl]);
							if (IS_NAN(p[sub][nucl]) || (p[sub][nucl] == numeric_limits<double>::infinity())) {
								stringstream ss;
								ss << "emission probability is nan/inf (sub: " << sub << ", pos: " << pos << ", nucl: " << nucl << ")";
								throw Error(ss.str());
							}							
							//cout << "p[" << sub << "][" << nucl << "]: " << p[sub][nucl] << " (" << exp(kappa + p_src[nucl]) << ")" << endl; 
						}
					//cout << endl;
				}
			}
		}	
		
		for (unsigned sub = 0; sub < subs.size(); ++sub) { 
			double sum = 0;
			for (unsigned nucl = 0; nucl < Consts::num_nucl(); ++nucl) {
				sum += p[sub][nucl];
			}
			for (unsigned nucl = 0; nucl < Consts::num_nucl(); ++nucl) {
				p[sub][nucl] /= sum;
				if (OUTPUT_DEBUG) 
					cout << "p[" << sub << "][" << nucl << "]: " << p[sub][nucl] << endl;
				p_em[sub][pos][nucl] = p[sub][nucl];
				if (IS_NAN(p[sub][nucl])) {
					stringstream ss;
					ss << "emission probability is nan (sub: " << sub << ", pos: " << pos << ", nucl: " << nucl << ")";
					throw Error(ss.str());
				}
			}			
		}
	}
	
	ofstream EP_file(EP_file_name().c_str());
	for (vector<unsigned>::iterator it = subs.begin(); it != subs.end(); ++it)
		for (unsigned j = 0; j < seq_len; ++j) {
			EP_file << j << " " << idx_to_sub_name[*it] << " ";
			for (unsigned k = 0; k < Consts::num_nucl(); ++k)
				EP_file << p_em[*it][j][k] << " ";
			EP_file  << endl;
		}
}



// calculate emission probabilities for strategy 'orig'
void calc_ep_orig(const mul_arr_int_3 em_freq, unsigned seq_len, set<unsigned> small_subs, set<unsigned> large_subs, double mu = 1)
{
	const unsigned MAX_SIZE_SMALL_SUB = 4;
	set<unsigned> subs_set(small_subs.begin(), small_subs.end());
	subs_set.insert(large_subs.begin(), large_subs.end());
	vector<unsigned> subs;
	copy(subs_set.begin(), subs_set.end(), back_inserter(subs));
	mul_arr_dbl_3 p_em;
	mul_arr_dbl_3::extent_gen extents;
	p_em.resize(extents[subs.size()][seq_len][Consts::num_nucl()]);

	for (vector<unsigned>::iterator it = subs.begin(); it != subs.end(); ++it) {
		unsigned size_sub = 0;
		for (unsigned j = 0; j < seq_len; ++j) {
			unsigned sum_freq = 0;
			for (unsigned k = 0; k < Consts::num_nucl(); ++k)
				sum_freq += em_freq[j][*it][k];
			size_sub = max(size_sub, sum_freq);
		}
		double scaling_factor = 1;
		if (size_sub <= MAX_SIZE_SMALL_SUB)
			scaling_factor = mu;
		for (unsigned j = 0; j < seq_len; ++j) {
			unsigned sum_freq = 0;
			for (unsigned k = 0; k < Consts::num_nucl(); ++k)
				sum_freq += em_freq[j][*it][k];
			for (unsigned k = 0; k < Consts::num_nucl(); ++k)
				p_em[*it][j][k] = (em_freq[j][*it][k] + scaling_factor * Consts::alpha()[k])
				 	 	 	 	  / (sum_freq + scaling_factor * Consts::sum_alpha());
		}
	}

	ofstream EP_file(EP_file_name().c_str());
	for (vector<unsigned>::iterator it = subs.begin(); it != subs.end(); ++it)
		for (unsigned j = 0; j < seq_len; ++j) {
			EP_file << j << " " << idx_to_sub_name[*it] << " ";
			for (unsigned k = 0; k < Consts::num_nucl(); ++k)
				EP_file << p_em[*it][j][k] << " ";
			EP_file  << endl;
		}
}



struct aic_para { unsigned pos; unsigned num_distr; vector<unsigned> sources; const mul_arr_int_3* em_freq; unsigned seq_len; vector<unsigned> subs; };

// auxiliary function for calc_aic_sources
double aic(const gsl_vector *v, void *params)
{
	const bool OUTPUT_DEBUG = 0;
	aic_para *para = (aic_para *)params;
	unsigned pos = para->pos;
	unsigned num_distr = para->num_distr;
	vector<unsigned> sources = para->sources;
	const mul_arr_int_3* em_freq = para->em_freq;
	vector<unsigned> subs = para->subs;
	
	mul_arr_dbl_2 p;
	mul_arr_dbl_2::extent_gen extents;
	p.resize(extents[num_distr][Consts::num_nucl()]);
  
	for (unsigned i = 0; i < num_distr; ++i) {
		double s = 0;
		for (unsigned j = 0; j < Consts::num_nucl(); ++j) {
			double x = exp(gsl_vector_get(v, i * Consts::num_nucl() + j));
			s += x;
			p[i][j] = x;
		}
		for (unsigned j = 0; j < Consts::num_nucl(); ++j)
			p[i][j] /= s;
	}
	
	if (OUTPUT_DEBUG) {
		for (unsigned i = 0; i < num_distr; ++i)
			for (unsigned j = 0; j < Consts::num_nucl(); ++j)
				cout << "p[" << i << "][" << j << "] = " << p[i][j] << endl;
	}
	
	double res = 0;
	for (unsigned j = 0; j < subs.size(); ++j) {
		for (unsigned k = 0; k < Consts::num_nucl(); ++k) {
			res += (*em_freq)[pos][j][k] * log(p[sources[j]][k]);
		}
	}
	res *= -1;
	
	if (OUTPUT_DEBUG)
		cout << "res: " << res << endl << endl;
	return res; 
}



// calculate AIC for different number of sources
void calc_aic_sources(const mul_arr_int_3 em_freq, unsigned seq_len, unsigned num_subs)
{
	const bool OUTPUT_PROGRESS = 1;
	const bool OUTPUT_DEBUG = 0;
	const bool OUTPUT_DEBUG_DETAILS = 0;
	const unsigned MAX_NUM_SOURCES = 100;
	unsigned max_num_sources = min(MAX_NUM_SOURCES, num_subs);
	
	vector<unsigned> subs;
	for (unsigned i = 0; i < num_subs; ++i)
		subs += i;
	
	vector<double> best_llhs(max_num_sources);
	vector< map<unsigned, set<unsigned> > > best_source_combs(max_num_sources);
	mul_arr_dbl_2 best_aic;
	mul_arr_dbl_2::extent_gen extents;
	best_aic.resize(extents[seq_len][max_num_sources]);
	vector<bool> pos_used(seq_len, false);
	
	cout << "calc_aic_sources" << endl;
	for (unsigned pos = 0; pos < seq_len; ++pos) {
		if (OUTPUT_PROGRESS)
			cout << "pos: " << pos << endl;
		bool enough_res = true;
		for (unsigned i = 0; i < num_subs; ++i) {
			bool b = false;
			for (unsigned j = 0; j < Consts::num_nucl(); ++j)
				if (em_freq[pos][i][j])
					b = true;
			enough_res &= b;
		}
			
		gsl_error_handler_t* gsl_error_handler = gsl_set_error_handler_off();
		if (enough_res) {
			if (OUTPUT_DEBUG) {
				for (unsigned i = 0; i < num_subs; ++i) {
					cout << "pos: "  << pos << "   sub: " << i << endl;
					for (unsigned j = 0; j < Consts::num_nucl(); ++j)
						cout << em_freq[pos][i][j] << " ";
					cout << endl;
				}
				cout << endl;
			}

			pos_used[pos] = true;
			for (unsigned num_sources = 1; num_sources <= max_num_sources; ++num_sources) {
				if (OUTPUT_DEBUG)
					cout << "num_sources: "  << num_sources << endl;
				double curr_best_llh = 0;
				map<unsigned, set<unsigned> > curr_best_source_comb;
				bool is_first = true;
				vector< map<unsigned, set<unsigned> > > source_combs;
				if (num_sources < 3)
					source_combs = get_all_source_partitions(num_sources, subs.size());
				else
					source_combs = get_derived_source_partitions(best_source_combs[num_sources - 2]);
				FOREACH (map<unsigned COMMA set<unsigned> > source_comb, source_combs) {
					vector<unsigned> source_vector = get_source_vector(source_comb);
					if (OUTPUT_DEBUG)
						cout << "source_comb: "  << source_comb << endl;
					const gsl_multimin_fminimizer_type *T = gsl_multimin_fminimizer_nmsimplex;
					gsl_multimin_fminimizer *s = NULL;
					gsl_vector *ss, *x;
					gsl_multimin_function minex_func;

					size_t iter = 0;
					int status;
					double size;
					int n = num_sources * Consts::num_nucl();

					x = gsl_vector_alloc(n);
					for (int i = 0; i < n; ++i)
						gsl_vector_set (x, i, 0.0);

					ss = gsl_vector_alloc(n);
					for (int i = 0; i < n; ++i)
						gsl_vector_set (ss, i, 1.0);

					minex_func.n = n;
					minex_func.f = &aic;
					aic_para par = { pos, num_sources, source_vector, &em_freq, seq_len, subs };
					minex_func.params = (void *)&par;

					s = gsl_multimin_fminimizer_alloc (T, n);
					gsl_multimin_fminimizer_set (s, &minex_func, x, ss);

					double curr_llh = 0;
					do {
						iter++;
						status = gsl_multimin_fminimizer_iterate(s);

						if (status)
							break;

						size = gsl_multimin_fminimizer_size (s);
						status = gsl_multimin_test_size (size, 1e-2);
						curr_llh = -s->fval;
						if (OUTPUT_DEBUG_DETAILS) {
							cout << "iter: " << iter << endl;
							cout << "size: " << size << endl;
							for (unsigned i = 0; i < num_sources; ++i) {
								double sum = 0;
								for (unsigned j = 0; j < Consts::num_nucl(); ++j)
									sum += exp(gsl_vector_get(s->x, i * Consts::num_nucl() + j));
								for (unsigned j = 0; j < Consts::num_nucl(); ++j)
									cout << "x[" << i << "][" << j << "] = "
										 << gsl_vector_get(s->x, i * Consts::num_nucl() + j)
										 << " (" << exp(gsl_vector_get(s->x, i * Consts::num_nucl() + j)) / sum << ")" << endl;
							}
							cout << "llh: " << curr_llh << endl;
						}
					}
					while (status == GSL_CONTINUE && iter < 1000);

					if ((is_first || (curr_llh > curr_best_llh))) {
						if ((status != GSL_CONTINUE) && (status != GSL_SUCCESS))
							cout << "Error in multidimensional minimization, position " << pos
							     << ", source_comb " << source_vector << endl;
						curr_best_llh = curr_llh;
						curr_best_source_comb = source_comb;
						is_first = false;
					}

					if (OUTPUT_DEBUG) {
						mul_arr_dbl_2 p;
						mul_arr_dbl_2::extent_gen extents;
						p.resize(extents[num_sources][Consts::num_nucl()]);

						for (unsigned i = 0; i < num_sources; ++i) {
							double sum = 0;
							for (unsigned j = 0; j < Consts::num_nucl(); ++j) {
								double y = exp(gsl_vector_get(s->x, i * Consts::num_nucl() + j));
								sum += y;
								p[i][j] = y;
							}
							for (unsigned j = 0; j < Consts::num_nucl(); ++j)
								p[i][j] /= sum;
						}

						for (unsigned i = 0; i < num_sources; ++i)
							for (unsigned j = 0; j < Consts::num_nucl(); ++j)
								cout << "p[" << i << "][" << j << "] = " << p[i][j] << endl;
						cout << endl;
					}

					gsl_vector_free(x);
					gsl_vector_free(ss);
					gsl_multimin_fminimizer_free (s);
				}
				best_llhs[num_sources - 1] = curr_best_llh;
				best_source_combs[num_sources - 1] = curr_best_source_comb;
				if (OUTPUT_DEBUG)
					cout << endl;
			}
			for (unsigned i = 0; i < max_num_sources; ++i)
				best_aic[pos][i] = best_llhs[i] - i * (Consts::num_nucl() - 1);
			if (OUTPUT_DEBUG) {
				for (unsigned i = 0; i < max_num_sources; ++i)
					cout << "best_aic[" << i << "] = " << best_aic[pos][i]
						 << " (" << get_source_vector(best_source_combs[i]) << ")" << endl;
			}
		}	
		gsl_set_error_handler(gsl_error_handler);

		/// todo: remove
		vector<int> freq_argmax(max_num_sources, 0);
		for (unsigned pos = 0; pos < seq_len; ++pos)
			if (pos_used[pos]) {
				double max = 0;
				int arg = 0;
				bool is_first = true;
				for (unsigned i = 0; i < max_num_sources; ++i) {
					if (is_first || (best_aic[pos][i] > max)) {
						is_first = false;
						max = best_aic[pos][i];
						arg = i;
					}
				}
				++freq_argmax[arg];
			}

		for (unsigned i = 0; i < max_num_sources; ++i)
			cout << "freq_argmax[" << i << "]: " << freq_argmax[i] << endl;


	}
	
	vector<int> freq_argmax(max_num_sources, 0);
	for (unsigned pos = 0; pos < seq_len; ++pos) 
		if (pos_used[pos]) {
			double max = 0;
			int arg = 0;
			bool is_first = true;
			for (unsigned i = 0; i < max_num_sources; ++i) {			
				if (is_first || (best_aic[pos][i] > max)) {
					is_first = false;
					max = best_aic[pos][i];
					arg = i;
				}			
			}
			++freq_argmax[arg];
		}
		
	for (unsigned i = 0; i < max_num_sources; ++i)
		cout << "freq_argmax[" << i << "]: " << freq_argmax[i] << endl;
}



int main(int argc, char** argv)
{	  
	const bool TEST_MANUAL = 1;
	if (argc == 1) {	
		if (TEST_MANUAL) {
			int seq_len;
			set<unsigned> small_subs, large_subs;
			mul_arr_int_3 em_freq;
			mul_arr_int_3::extent_gen extents;
			em_freq.resize(extents[1][9][Consts::num_nucl()]);

			em_freq[0][0][0] = 0;
			em_freq[0][0][1] = 10;

			em_freq[0][1][0] = 0;
			em_freq[0][1][1] = 10;

			em_freq[0][2][0] = 10;
			em_freq[0][2][1] = 0;

			
			seq_len = 1;
			small_subs += 0;
			small_subs += 1;
			small_subs += 2;
				
			calc_aic_sources(em_freq, seq_len, small_subs.size());
			//calc_ep(em_freq, seq_len, small_subs, large_subs);	
			return 0;
		}
		else {
			//string cmd_line(" /gobics/home/ingo/workspace/calcETP/Debug/calcETP --em --tr 2009-03-09_11-12-30_5666 ");
			string cmd_line(" /home/ingo/workspace/small_subtypes/Debug/calc_ETP --em --tr --cons --al /home/ingo/tmp/small_subtypes/calcETP_al/alignment_simple.fas --stamp 2010-06-04_11-20-38_8084 --home /home/ingo ");
			boost::regex e("\\s*(\\S.*\\S)\\s*");
			cmd_line = regex_replace(cmd_line, e, "\\1", boost::match_default | boost::format_sed);
			
			vector<string> v; 
		    split(v, cmd_line, is_any_of(" "));
		    
		    argc = v.size();
		    argv = new char*[argc];
		    for (unsigned i = 0; i < v.size(); ++i) {
		    	argv[i] = new char[v[i].length() + 1];	    
		    	strcpy(argv[i], v[i].c_str());
		    }
		}
	}
	
	options_description desc("Allowed options");
	desc.add_options()
		("em_prob", "calculate emission probabilities using strategy 'prob'")
		("em_orig", "calculate emission probabilities using strategy 'orig'")
		("em_pc_scaling", "calculate emission probabilities using strategy 'orig', scaling pseudocounts for small subtypes")
		("mu", value<double>(), "scaling factor for pseudocounts")
		("tr", "calculate transition probabilities")
		("cons", "calculate consensus columns")
		("repr", "calculate representative emission probabilities")
		("sources", "calculate AIC for different number of sources")
		("stamp", value<string>(), "stamp attached to output filenames")
		("scratch", "write output filenames directly to scratch")
		("al", value<string>(), "alignment file name")
		("home", value<string>(), "home directory name");

	positional_options_description pod;
	variables_map vm;
	store(parse_command_line(argc, argv, desc), vm);
	notify(vm);
	
	mp::mp_init(300);

	// save stamp
	if (vm.count("stamp")) {
		Consts::stamp() = vm["stamp"].as<string>();
	}
	else
		throw Error("stamp not provided as argument");

	if (vm.count("al"))
		alignment_file_name() = vm["al"].as<string>();
	else
		throw Error("alignment file name not provided as argument");

	if (vm.count("home"))
		home_dir_name() = vm["home"].as<string>();
	else
		throw Error("home directory name not provided as argument");

	Consts::use_scratch() = vm.count("scratch");

    insert( nucl_to_idx )
		( 'A', 0 )
		( 'C', 1 )
		( 'G', 2 )
		( 'T', 3 );
	get_alignment();
    
	const int NUM_DIR_MIX_EM = 2;
	const int NUM_DIR_MIX_MATCH = 1;
	
	// calculate AIC for different number of sources
	if (vm.count("sources")) {
		unsigned seq_len, num_subs;
		mul_arr_int_3 em_freq;
		calc_repr_em_freq(&em_freq, &seq_len, &num_subs);
	
		/*mul_arr_int_3::extent_gen extents;
		em_freq.resize(extents[1][3][consts::num_nucl()]);
		em_freq[0][0][0] = 50;
		em_freq[0][0][1] = 1;
		em_freq[0][0][2] = 0;
		em_freq[0][0][3] = 0;
		em_freq[0][1][0] = 0;
		em_freq[0][1][1] = 50;
		em_freq[0][1][2] = 0;
		em_freq[0][1][3] = 0;
		em_freq[0][2][0] = 0;
		em_freq[0][2][1] = 1;
		em_freq[0][2][2] = 0;
		em_freq[0][2][3] = 0;
		
		seq_len = 1;
		num_subs = 3;*/
			
		calc_aic_sources(em_freq, seq_len, num_subs);		
	}
	
	// calculate representative emissions and transition probabilities
	if (vm.count("repr")) {
		// calculate representative emission probabilities
		unsigned num_pos_repr;
		mul_arr_int_2 repr_em_freq;
		get_repr_em_freq(&repr_em_freq, &num_pos_repr);
	
		mul_arr_dbl_2 alpha_em(boost::extents[NUM_DIR_MIX_EM][Consts::num_nucl()]);
		vector<double> q_em(NUM_DIR_MIX_EM);
		est_para_dir_mix(repr_em_freq, Consts::num_nucl(), NUM_DIR_MIX_EM, num_pos_repr, &alpha_em, &q_em);
		save_repr_em_para(NUM_DIR_MIX_EM, alpha_em, q_em);
		cout << "representative emission probabilities calculated" << endl;
		
		// calculate representative transition probabilities		
		/*mul_arr_int_2 repr_tr_freq;
		get_repr_tr_freq(&repr_tr_freq, &num_pos_repr);

		mul_arr_dbl_2 alpha_match(boost::extents[NUM_DIR_MIX_MATCH][2]);
		vector<double> q_match(NUM_DIR_MIX_MATCH);		
		est_para_dir_mix(repr_tr_freq, 2, NUM_DIR_MIX_MATCH, num_pos_repr, &alpha_match, &q_match);		
		save_repr_tr_para(NUM_DIR_MIX_MATCH, alpha_match, q_match);	
		cout << "representative transition probabilities calculated" << endl;*/
	}

	// calculate emission probabilities for strategy 'prob'
	if (vm.count("em_prob")) {
		unsigned seq_len;
		set<unsigned> small_subs, large_subs;
		mul_arr_int_3 em_freq;
		calc_em_freq(&em_freq, &seq_len, &small_subs, &large_subs);
	
		/*mul_arr_int_3::extent_gen extents;
		em_freq.resize(extents[1][3][consts::num_nucl()]);
		em_freq[0][0][0] = 10;
		em_freq[0][0][1] = 0;
		//em_freq[0][0][2] = 15;
		//em_freq[0][0][3] = 0;
		em_freq[0][1][0] = 1;
		em_freq[0][1][1] = 10;
		//em_freq[0][1][2] = 1;
		//em_freq[0][1][3] = 0;
		em_freq[0][2][0] = 1;
		em_freq[0][2][1] = 0;
		//em_freq[0][2][2] = 1;
		//em_freq[0][2][3] = 0;
		
		seq_len = 1;
		small_subs += 0;
		small_subs += 1;
		small_subs += 2;*/
			
		calc_ep_prob(em_freq, seq_len, small_subs, large_subs);
	}

	// calculate emission probabilities for strategy 'orig'
	if (vm.count("em_orig") || vm.count("em_pc_scaling")) {
		unsigned seq_len;
		set<unsigned> small_subs, large_subs;
		mul_arr_int_3 em_freq;
		calc_em_freq(&em_freq, &seq_len, &small_subs, &large_subs);

		if (vm.count("em_orig"))
			calc_ep_orig(em_freq, seq_len, small_subs, large_subs);
		else {
			if (!vm.count("mu"))
				throw Error("mu not given");
			double mu = vm["mu"].as<double>();
			calc_ep_orig(em_freq, seq_len, small_subs, large_subs, mu);
		}
	}

	// calculate transition probabilities
	if (vm.count("tr")) {		
		unsigned num_dir_mix_match;
		mul_arr_dbl_2 alpha_match(boost::extents[NUM_DIR_MIX_MATCH][2]);
		vector<double> q_match(NUM_DIR_MIX_MATCH);
		load_repr_tr_para(&num_dir_mix_match, &alpha_match, &q_match);
		
		unsigned seq_len;
		set<unsigned> small_subs, large_subs;
		mul_arr_int_3 tr_freq;
		calc_tr_freq(&tr_freq, &seq_len, &small_subs, &large_subs);
		
		unsigned num_subs = small_subs.size() + large_subs.size();		
		mul_arr_int_2 freq_match(boost::extents[seq_len - 1][2]);
		for (mul_arr_int_3::index i = 0; i < seq_len - 1; ++i)
			for (mul_arr_int_3::index j = 0; j < 2; ++j)
				for (mul_arr_int_3::index k = 0; k < num_subs; ++k)
					freq_match[i][j] += tr_freq[i][k][j];
				
		mul_arr_int_2 freq_del(boost::extents[seq_len - 1][2]);
		for (mul_arr_int_3::index i = 0; i < seq_len - 1; ++i)
			for (mul_arr_int_3::index j = 0; j < 2; ++j)
				for (mul_arr_int_3::index k = 0; k < num_subs; ++k)
					freq_del[i][j] += tr_freq[i][k][j + 2];
		
		mul_arr_dbl_2 tr_prob(boost::extents[seq_len - 1][Consts::num_tr_probs()]);
		for (mul_arr_dbl_2::index i = 0; i < seq_len - 1; ++i) {
			vector<double> p(2);
			vector<int> n(2);
			n[0] = freq_match[i][0];
			n[1] = freq_match[i][1];
			int sum_n = n[0] + n[1];
			est_prob_from_dir_mix(2, NUM_DIR_MIX_MATCH, n, alpha_match, q_match, &p);
			tr_prob[i][0] = p[0] * sum_n / (sum_n + 1) + 0.8 / (sum_n + 1);
			tr_prob[i][1] = 0.1 / (sum_n + 1);
			tr_prob[i][2] = p[1] * sum_n / (sum_n + 1) + 0.1 / (sum_n + 1);
	
			tr_prob[i][3] = 0.5;
			tr_prob[i][4] = 0.5;
	
			n[0] = freq_del[i][0];
			n[1] = freq_del[i][1];
			sum_n = n[0] + n[1];
			tr_prob[i][5] = (n[0] + 0.25) / (sum_n + 0.5);
			tr_prob[i][6] = (n[1] + 0.25) / (sum_n + 0.5);
		}
		
		ofstream TP_file(TP_file_name().c_str());
		for (unsigned i = 0; i < num_subs; i++) {
			for (unsigned j = 0; j < seq_len - 1; j++) {
				TP_file << j << " " << idx_to_sub_name[i] << " ";
				for (unsigned k = 0; k < Consts::num_tr_probs(); k++)
					TP_file << tr_prob[j][k] << " ";
				TP_file << endl;
			}
	
			TP_file << seq_len - 1 << " " << i << " ";
			for (unsigned k = 0; k < Consts::num_tr_probs(); k++)
				TP_file << 0 << " ";
			TP_file << endl;
		}
	}
	
	if (vm.count("cons")) {
		unsigned seq_len;
		set<unsigned> small_subs, large_subs;
		mul_arr_int_3 em_freq;
		calc_em_freq(&em_freq, &seq_len, &small_subs, &large_subs);
		set<int> subs(small_subs.begin(), small_subs.end());
		subs.insert(large_subs.begin(), large_subs.end());
		
		ofstream cons_file(cons_file_name().c_str());
		for (unsigned pos = 0; pos < seq_len; ++pos) {
			bool only_gaps = true;
			FOREACH (int sub, subs)
				for (unsigned nucl = 0; nucl < Consts::num_nucl(); ++nucl)
					if (em_freq[pos][sub][nucl])
						only_gaps = false;
			if (!only_gaps)
				cons_file << pos << endl;
		}			
	}
	
	cout << "Calc_ETP finished with success" << endl;

	return 1;
}

