#include "consts.h"
#include "dirichlet.h"
#include "error.h"

#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <deque>
#include <map>

#include <gsl/gsl_multimin.h>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_roots.h>
#include <arprec/mp_real.h>

using namespace std;
using namespace boost;



// calculates lgo P(p|α) (cf. Sjölander (1996), Lemma 2)
double log_prob_p_given_alpha(vector<double> p, vector<double> alpha)
{
	double ret = 0;
	double sum_alpha = accumulate(alpha.begin(), alpha.end(), 0.0);
	double r = gsl_sf_gamma(sum_alpha);
	for (unsigned i = 0; i < Consts::num_nucl(); ++i)
		r -= gsl_sf_gamma(alpha[i]);
	for (unsigned i = 0; i < Consts::num_nucl(); ++i)
		ret += p[i] * (alpha[i] - 1);
	return ret;
}


// likelihood of a parameter choice for a Dirichlet mixture (cf. Sjölander (1996), p. 343, (A.3))
double lh_para_dir_mix(const gsl_vector *v, void *params)
{
	const bool OUTPUT_DEBUG = 0;

	const int dim = ((lh_para_dir_mix_para*)params)->dim;
	const int num_dir_mix = ((lh_para_dir_mix_para*)params)->num_dir_mix;
	const int num_pos = ((lh_para_dir_mix_para*)params)->num_pos;
	mul_arr_int_2 n = ((lh_para_dir_mix_para*)params)->n;

	if (OUTPUT_DEBUG) {
		cout << "dim: " << dim << endl;
		cout << "num_dir_mix: " << num_dir_mix << endl;
		cout << "num_pos: " << num_pos << endl;
		for (int i = 0; i < num_pos; ++i)
			for (int j = 0; j < dim; ++j)
				cout << "n[" << i << "][" << j << "]: " << n[i][j] << endl;
	}

	deque<double> q(num_dir_mix);
	mul_arr_dbl_2 alpha(boost::extents[num_dir_mix][dim]);
    for (int i = 0; i < num_dir_mix; ++i)
    	for (int j = 0; j < dim; ++j)
    		alpha[i][j] = exp(gsl_vector_get(v, i * dim + j));
    q[0] = 1;
    for (int i = 0; i < num_dir_mix - 1; ++i)
    	q[i + 1] = exp(gsl_vector_get(v, num_dir_mix * dim + i));

	double sum_q = 0;
	for (int i = 0; i < num_dir_mix; ++i)
		sum_q += q[i];
	for (int i = 0; i < num_dir_mix; ++i)
		q[i] /= sum_q;

	if (OUTPUT_DEBUG) {
		for (int i = 0; i < num_dir_mix; ++i)
			for (int j = 0; j < dim; ++j)
				cout << "alpha[" << i << "][" << j << "]: " << alpha[i][j] << endl;
		for (int i = 0; i < num_dir_mix; ++i)
				cout << "q[" << i << "]: " << q[i] << endl;
		cout << endl;
	}

	mp_real ret = 1;
	for (int i = 0; i < num_pos; ++i) {
		int sum_n = 0;
		for (int j = 0; j < dim; ++j)
			sum_n += n[i][j];

		mp_real p_pos = 0.0;
		for (int j = 0; j < num_dir_mix; ++j) {
			double sum_alpha = 0;
			for (int k = 0; k < dim; ++k)
				sum_alpha += alpha[j][k];
			double p = gsl_sf_lngamma(sum_n + 1) + gsl_sf_lngamma(sum_alpha) - gsl_sf_lngamma(sum_n + sum_alpha);
			for (int k = 0; k < dim; ++k) {
				p += gsl_sf_lngamma(n[i][k] + alpha[j][k]) - gsl_sf_lngamma(n[i][k] + 1) - gsl_sf_lngamma(alpha[j][k]);
			}
			p_pos += q[j] * exp((mp_real)p);
		}
		ret *= p_pos;
	}
	
	return dble(-log(ret));
}



// estimate parameters of Dirichlet mixture
void est_para_dir_mix(const mul_arr_int_2 freq, const int dim, const int num_dir_mix, const int num_pos, mul_arr_dbl_2* alpha, vector<double>* q)
{
	const bool OUTPUT_DEBUG = 1;
	const double PRECISION = 1e-2;

	const gsl_multimin_fminimizer_type *T =  gsl_multimin_fminimizer_nmsimplex;
	gsl_multimin_fminimizer *s = NULL;
	gsl_vector *ss, *x;
	gsl_multimin_function minex_func;

	size_t iter = 0;
	int status;
	
	gsl_rng* rng = gsl_rng_alloc(gsl_rng_mt19937);
	int rng_seed = time(0);
	gsl_rng_set(rng, rng_seed);
	  
	x = gsl_vector_alloc (num_dir_mix * dim + num_dir_mix - 1);
	for (int i = 0; i < num_dir_mix * dim + num_dir_mix - 1; ++i)
		gsl_vector_set (x, i, gsl_rng_uniform(rng));

	ss = gsl_vector_alloc (num_dir_mix * dim + num_dir_mix - 1);
	gsl_vector_set_all (ss, 1.0 + gsl_rng_uniform(rng));

	minex_func.n = num_dir_mix * dim + num_dir_mix - 1;
	minex_func.f = &lh_para_dir_mix;
	lh_para_dir_mix_para par;
	par.dim = dim;
	par.num_dir_mix = num_dir_mix;
	par.num_pos = num_pos;
	mul_arr_int_2::extent_gen extents;
	par.n.resize(extents[num_pos][dim]);
	par.n = freq;
	minex_func.params = (void *)&par;

	s = gsl_multimin_fminimizer_alloc (T, num_dir_mix * dim + num_dir_mix - 1);
	gsl_multimin_fminimizer_set (s, &minex_func, x, ss);

	bool solution_singular = false;
	do
	{
		iter++;
	    status = gsl_multimin_fminimizer_iterate(s);

	    if (status)
	    	break;

	    double size = gsl_multimin_fminimizer_size (s);
	    status = gsl_multimin_test_size (size, PRECISION);

	    if (OUTPUT_DEBUG) {
	    	cout << "iter: " << iter << endl;
			cout << "alpha:" << endl;
			for (int i = 0; i < num_dir_mix; ++i) {
				for (int j = 0; j < dim; ++j)
					cout << exp(gsl_vector_get (s->x, i * dim + j)) << " ";
				cout << endl;
			}
			cout << "q: 1 ";
			for (int i = 0; i < num_dir_mix - 1; ++i)
				cout << exp(gsl_vector_get (s->x, num_dir_mix * dim + i)) << " ";
			cout << endl;
			cout << "size: " << size << endl;
			cout << "llh: " << s->fval << endl << endl;
	    }

	    const double LARGE = 1e10;
	    const double SMALL = 1e-10;
		for (int i = 0; i < num_dir_mix; ++i)
			for (int j = 0; j < dim; ++j)
				if ((exp(gsl_vector_get (s->x, i * dim + j)) > LARGE) ||
					(exp(gsl_vector_get (s->x, i * dim + j)) < SMALL))
					solution_singular = true;
	}
	while ((status == GSL_CONTINUE) && !solution_singular);

	for (int i = 0; i < num_dir_mix; ++i)
		for (int j = 0; j < dim; ++j)
			(*alpha)[i][j] = exp(gsl_vector_get(s->x, i * dim + j));
	(*q)[0] = 1;
	for (int i = 0; i < num_dir_mix - 1; ++i)
		(*q)[i + 1] = exp(gsl_vector_get(s->x, num_dir_mix * dim + i));
	double sum_q = 0;
	for (int i = 0; i < num_dir_mix; ++i)
		sum_q += (*q)[i];
	for (int i = 0; i < num_dir_mix; ++i)
		(*q)[i] /= sum_q;

	gsl_vector_free(x);
	gsl_vector_free(ss);
	gsl_multimin_fminimizer_free (s);
}



// calculate probability of frequencies n given alpha and |n|
mp_real prob_n_given_dir_mix(const int dim, const int num_dir_mix, const vector<int> n,
						   	 const mul_arr_dbl_2 alpha, const vector<double> q)
{
	mp_real ret = 0.0;
	int sum_n = 0;
	for (int i = 0; i < dim; ++i)
		sum_n += n[i];

	for (int i = 0; i < num_dir_mix; ++i) {
		double sum_alpha = 0;
		for (int j = 0; j < dim; ++j)
			sum_alpha += alpha[i][j];
		double p = gsl_sf_lngamma(sum_n + 1) + gsl_sf_lngamma(sum_alpha) - gsl_sf_lngamma(sum_n + sum_alpha);
		for (int j = 0; j < dim; ++j) {
			p += gsl_sf_lngamma(n[j] + alpha[i][j]) - gsl_sf_lngamma(n[j] + 1) - gsl_sf_lngamma(alpha[i][j]);
		}
		ret += q[i] * exp((mp_real)p);
	}

	return ret;
}



// estimate probability (p_1, ..., p_dim) given a Dirichlet mixture (cf. Sjölander (1996), p. 335, (15))
void est_prob_from_dir_mix(const int dim, const int num_dir_mix, const vector<int> n,
						   const mul_arr_dbl_2 alpha, const vector<double> q, vector<double>* p)
{
	const bool OUTPUT_DEBUG = 0;

	// calculate probability of alpha_j given n and Theta (eq. (16))
	vector<double> prob_alpha;
	double sum_prob_alpha = 0;
	for (int i = 0; i < num_dir_mix; ++i) {
		vector<double> curr_q(1, 1);
		mul_arr_dbl_2 curr_alpha(boost::extents[1][dim]);
		for (int j = 0; j < dim; ++j)
			curr_alpha[0][j] = alpha[i][j];
		double x = q[i] * dble(prob_n_given_dir_mix(dim, 1, n, curr_alpha, curr_q));
		prob_alpha.push_back(x);
		sum_prob_alpha += x;
	}
	for (int i = 0; i < num_dir_mix; ++i)
		prob_alpha[i] /= sum_prob_alpha;

	double sum_n = 0;
	for (int i = 0; i < dim; ++i)
		sum_n += n[i];

	vector<double> sum_alpha;
	for (int i = 0; i < num_dir_mix; ++i) {
		double x = 0;
		for (int j = 0; j < dim; ++j)
			x += alpha[i][j];
		sum_alpha.push_back(x);
	}

	for (int i = 0; i < dim; ++i) {
		(*p)[i] = 0;
		for (int j = 0; j < num_dir_mix; ++j)
			(*p)[i] += prob_alpha[j] * (n[i] + alpha[j][i]) / (sum_n + sum_alpha[j]);
	}

	if (OUTPUT_DEBUG) {
		for (int i = 0; i < num_dir_mix; ++i)
			cout << "prob_alpha[" << i << "]: " << prob_alpha[i] << endl;
		for (int i = 0; i < num_dir_mix; ++i)
			for (int j = 0; j < dim; ++j)
				cout << "alpha[" << i << "][" << j << "]: " << alpha[i][j] << endl;
		for (int i = 0; i < num_dir_mix; ++i)
			cout << "sum_alpha[" << i << "]: " << sum_alpha[i] << endl;
		for (int i = 0; i < dim; ++i)
			cout << "n[" << i << "]: " << n[i] << endl;
		cout << "sum_n: " << sum_n << endl;
		for (int i = 0; i < dim; ++i)
			cout << "p[" << i << "]: " << (*p)[i] << endl;
		cout << endl;
	}
}


double aux_inv_ln_gamma (double y, void *params) 
{ 
	double* x = (double*) params;
	return gsl_sf_lngamma(y) - *x; 
}


// calculates y in [x_lo, x_hi] such that Gamma(y) = x
double inv_ln_gamma(double x, double x_lo, double x_hi)
{
	int status;
	int iter = 0, max_iter = 100;
	const gsl_root_fsolver_type *T;
	gsl_root_fsolver *s;
	double r = 0;
	double eps = 1e-10;
	gsl_function F;
	
	F.function = &aux_inv_ln_gamma;
	F.params = &x;
	
	T = gsl_root_fsolver_brent;
	s = gsl_root_fsolver_alloc (T);
	gsl_root_fsolver_set (s, &F, x_lo, x_hi);
	
	do {
	    iter++;
	    status = gsl_root_fsolver_iterate (s);
	    r = gsl_root_fsolver_root (s);
	    x_lo = gsl_root_fsolver_x_lower (s);
	    x_hi = gsl_root_fsolver_x_upper (s);
	    status = gsl_root_test_interval (x_lo, x_hi, 0, eps);
	} while (status == GSL_CONTINUE && iter < max_iter);
	
	gsl_root_fsolver_free (s);
	
	return (x_lo + x_hi) / 2;
}



pair<double, double> inv_ln_gamma(double x) 
{
	const double MIN_GAMMA = 1.461632144968362;
	return make_pair(inv_ln_gamma(x, 1e-100, MIN_GAMMA), inv_ln_gamma(x, MIN_GAMMA, 1e100)); 
}
