% a subtype segmentation
classdef Segmentation
    properties
        begin_pos
        end_pos
        bps
        subs
    end
    
    properties (Dependent)
        len
        num_bps
        num_segms
        used_subs
    end
    
    methods
        function obj = Segmentation(begin_pos, end_pos, bps, subs, offset)
            obj.begin_pos = begin_pos;
            obj.end_pos = end_pos;
            obj.bps = bps + offset;
            obj.subs = subs;
        end
        
        function len = get.len(obj)
            len = obj.end_pos - obj.begin_pos + 1;
        end
        function num_bps = get.num_bps(obj)
            num_bps = length(obj.bps);
        end
        function num_segms = get.num_segms(obj)
            num_segms = obj.num_bps + 1;
        end        
        function used_subs = get.used_subs(obj)
            used_subs = sort(unique(obj.subs));
        end
   
        % get sub at given seq position 
        function sub = sub_at_pos(obj, pos)
            i = 1;
            while (i <= obj.num_bps) && (pos > obj.bps(i)) 
                i = i + 1;
            end
            if (i > obj.num_bps)
                sub = obj.subs{end};
            else
                sub = obj.subs{i};
            end
        end
        
        % score bps and subs of segms of other segmentation
        function res = score(obj, other_segm, test_seq)  
            res = struct('score', 0, 'pos_score', 0, 'segm_score', 0);
            FRACTION_POS = 0.5;
            FRACTION_SEGM = 0.5;
            if (obj.begin_pos ~= other_segm.begin_pos) || (obj.end_pos ~= other_segm.end_pos)
                error 'trying to compare segmentations with different begin / end pos'
            end
            
            % calculate positionwise mismatches
            pos_score = 0;
            len_without_gaps = 0;
            for i=obj.begin_pos:obj.end_pos
                if ~Alpha.is_gap(test_seq.genome(i - obj.begin_pos + 1))
                    len_without_gaps = len_without_gaps + 1;
                    if obj.sub_at_pos(i) == other_segm.sub_at_pos(i)
                        pos_score = pos_score + 1;
                    end
                end
            end
            res.pos_score = pos_score / len_without_gaps;
            
            res.segm_score = (obj.segm_score(other_segm, test_seq) + other_segm.segm_score(obj, test_seq)) / 2;            
            res.score = FRACTION_SEGM * res.segm_score + FRACTION_POS * res.pos_score;
        end
        
        % get the segmentation score
        function segm_score = segm_score(obj, other_segm, test_seq)
            CUT_OFF = 50;
            segm_score = 0;
            for i=1:obj.num_segms
                % calculate begin and end pos of current segm
                if i == 1
                    curr_segm_begin = obj.begin_pos;
                else
                    curr_segm_begin = obj.bps(i - 1);
                end
                if i == obj.num_segms
                    curr_segm_end = obj.end_pos;
                else
                    curr_segm_end = obj.bps(i) - 1;
                end
                
                % check whether sub of current segm matches roughly with
                % other segmentation
                begin_wo_gaps = test_seq.pos_without_gaps(curr_segm_begin - obj.begin_pos + 1);
                end_wo_gaps = test_seq.pos_without_gaps(curr_segm_end - obj.begin_pos + 1);
                mid_point = round((begin_wo_gaps + end_wo_gaps) / 2);     
                is_match = true;
                for j=min(begin_wo_gaps + CUT_OFF, mid_point):max(end_wo_gaps - CUT_OFF, mid_point)
                    if obj.sub_at_pos(obj.begin_pos + 1 + test_seq.pos_with_gaps(j)) ~= other_segm.sub_at_pos(obj.begin_pos + 1 + test_seq.pos_with_gaps(j))
                        is_match = false;
                    end
                end
                if is_match
                    segm_score = segm_score + 1;
                end                
            end
            segm_score = segm_score / obj.num_segms;
        end
    end    
        
    methods (Static)
        function str = segm2str(segm, detailed)
            if detailed  
                % determine how many nucleotides correspond to one printed
                % character
                POSSIBLE_LEN = [1 2 5 10 20 50 100 200 500 1000];
                MAX_LEN = 160;
                i = 1;
                while (segm.len / POSSIBLE_LEN(i) > MAX_LEN)
                    i = i + 1;
                end
                len = POSSIBLE_LEN(i);
                str = ['(' num2str(len) ')'];
                
                % determine string
                j = 0;
                for i=segm.begin_pos:segm.end_pos
                    if ~j
                        curr_sub = segm.sub_at_pos(i);
                        is_mixed = false;
                    else
                        if ~is_mixed && (strcmp(segm.sub_at_pos(i), curr_sub))
                            is_mixed = true;
                        end
                    end
                    if (j == len - 1) || (i == segm.end_pos)
                        if is_mixed
                            str = [str '*'];                            
                        else
                            str = [str curr_sub];
                        end
                        j = 0;
                    else
                        j = j + 1;
                    end
                end
            else
                str = ['(' num2str(segm.begin_pos) ') '];
                for j=1:length(segm.subs)
                    str = [str segm.subs{j} ' '];
                    if j < length(segm.subs)
                        str = [str num2str(segm.bps(j)) ' '];
                    else
                        str = [str '(' num2str(segm.end_pos) ')'];
                    end
                end                 
            end
        end
    end
    
end