% a single test sequence in TestSeqs
classdef TestSeq
    properties
        begin_pos
        end_pos
        segm
        no_seqs_of_segms
        genome
        al
        pos_with_gaps
        pos_without_gaps
    end
    properties (SetAccess = private, GetAccess = private)
        type
    end
    properties (Dependent = true)
        bps
        num_bps
        num_segms
        used_subs
        segm_subs
        no_used_seqs
        len
        is_sens
        is_spec
    end
    
    methods
        % Constructor
        % begin_pos - the first position in the alignment used for the
        % sequence 
        % end_pos - the last position in the alignment used for the
        % sequence
        % bps - the breakpoints of the sequence
        % no_seqs_of_segms
        % type - the type of the sequence: 'spec' or 'sens'
        % al - the alignment
        function s = TestSeq(begin_pos, end_pos, bps, no_seqs_of_segms, type, al)
            % save begin and end pos
            if ((begin_pos >= end_pos) || (begin_pos < 1) || (end_pos > al.seq_len()))
                error(['begin_pos and end_pos have to fulfill 1 <= ' begin_pos ' < ' end_pos ' <= ' al.seq_len])
            end
            s.begin_pos = begin_pos;
            s.end_pos = end_pos;
            
            % save numbers of seqs
            for i=1:length(no_seqs_of_segms)
                if ((no_seqs_of_segms(i) < 1) || (no_seqs_of_segms(i) > al.num_seqs))
                    error(['The numbers of the seqs are too low/high (min: 1, max: ' al.num_seqs ', actual: ' no_segm_seqs(i) ')'])
                end
            end
            s.no_seqs_of_segms = no_seqs_of_segms;

            % save segm
            for i=2:length(bps)
                if (bps(i) <= bps(i - 1))
                    error('BPs are not strictly increasing')
                end
                if (bps(1) < begin_pos)
                    error('First BP too small')
                end
                if  (bps(end) > end_pos - 1)
                    error('Last BP too large')
                end
            end
            subs = {};
            for i=1:length(s.no_seqs_of_segms)
                subs = [subs al.seqs(s.no_seqs_of_segms(i)).sub];
            end
            s.segm = Segmentation(begin_pos, end_pos, bps, subs, 0);
            
            % generate genome
            if ~isempty(bps) 
                s.genome = al.seqs(no_seqs_of_segms(1)).genome(begin_pos:bps(1));
                for i=2:length(no_seqs_of_segms)-1
                    s.genome = [s.genome al.seqs(no_seqs_of_segms(i)).genome(bps(i-1)+1:bps(i))];
                end
                s.genome = [s.genome al.seqs(no_seqs_of_segms(end)).genome(bps(end)+1:end_pos)];

            else
                s.genome = al.seqs(no_seqs_of_segms(1)).genome(begin_pos:end_pos);
            end
            
            % get type
            switch type
                case 'sens'
                    s.type = 'sens';
                case 'spec'
                    s.type = 'spec';
                otherwise
                    error 'type of test seq has to be sens of spec'
            end

            % positions with and without gaps
            s.pos_with_gaps = [];
            s.pos_without_gaps = [];
            n = 0;
            for i=1:length(s.genome)
                if ~Alpha.is_gap(s.genome(i))
                    n = n + 1;
                    s.pos_with_gaps(n) = i;
                end
                s.pos_without_gaps(i) = n;
            end
            s.al = al;
        end
        
        function bps = get.bps(obj)
            bps = obj.segm.bps;
        end
        function num_bps = get.num_bps(obj)
            num_bps = obj.segm.num_bps;
        end
        function num_segms = get.num_segms(obj)
            num_segms = obj.segm.num_segms;
        end        
        function used_subs = get.used_subs(obj)
            used_subs = obj.segm.used_subs;
        end
        function segm_subs = get.segm_subs(obj)
            segm_subs = obj.segm.subs;
        end   
        function no_used_seqs = get.no_used_seqs(obj)
            no_used_seqs = sort(unique(obj.no_seqs_of_segms));
        end
        function len = get.len(obj)
            len = obj.end_pos - obj.begin_pos + 1;
        end
        function is_spec = get.is_spec(obj)
            is_spec = (obj.type == 'spec');
        end
        function is_sens = get.is_sens(obj)
            is_sens = (obj.type == 'sens');
        end
        
        % get sub at given seq position 
        function sub = sub_at_pos(obj, pos)
            sub = obj.segm.sub_at_pos(pos);
        end
        
        % apply jpHMM to seq and return score
        function job = submit_score_job(obj, model_para, al)
            global consts;
            global scenario;
            global manual_cluster;
            stamp = Stamp();
                        
            al.print_jpHMM_al(obj.begin_pos, obj.end_pos, obj.no_used_seqs, stamp);
            obj.print_jpHMM_query(stamp);
            
            switch consts.strategy
                case 'orig'
                    if (~isa(model_para, 'OrigPara'))
                        error 'Model para not of type OrigPara'
                    end
                    
                    % call calcETP
                    obj.print_calc_etp_seqs(al, stamp);   
                    if (model_para.mu == 1)
                        cmd = [FilesDirs.calc_etp_dir '/calc_ETP --em_orig --tr --cons --al ' ...
                               FilesDirs.calc_etp_al_filename(scenario) ' --stamp ' stamp ...
                               ' --home ' FilesDirs.home_dir()]  
                    else
                         cmd = [FilesDirs.calc_etp_dir '/calc_ETP --em_pc_scaling --tr --cons --al ' ...
                                FilesDirs.calc_etp_al_filename(scenario) ' --mu ' num2str(model_para.mu) ' --stamp ' stamp ...
                               ' --home ' FilesDirs.home_dir()]  
                    end
                    if consts.use_scratch
                        cmd = [cmd ' --scratch'];
                    end
                    calc_etp_job = Job(consts.use_cluster, true, 'calcETP', stamp, cmd, []);
                    calc_etp_job.submit();

                    % call jpHMM
                    jphmm_job = JPHMM.create_job(model_para, stamp, []);                    
                    jphmm_job.submit;                    
                    job = jphmm_job;                    
                case 'prob'
                    if (~isa(model_para, 'BasicPara'))
                        error 'Model para not of type ModSelPara'
                    end

                    % call calcETP
                    obj.print_calc_etp_seqs(al, stamp);                    
                    cmd = [FilesDirs.calc_etp_dir '/calc_ETP --em_prob --tr --cons --al ' FilesDirs.calc_etp_al_filename(scenario) ' --stamp ' stamp ...
                           ' --home ' FilesDirs.home_dir()]
                    if consts.use_scratch
                        cmd = [cmd ' --scratch'];
                    end
                    calc_etp_job = Job(consts.use_cluster, true, 'calcETP', stamp, cmd, []);
                    calc_etp_job.submit();
                                        
                    % call jpHMM
                    jphmm_job = JPHMM.create_job(model_para, stamp, calc_etp_job.cluster_id);
                    jphmm_job.submit;
                    job = jphmm_job;
                case 'ml'
                    if (~isa(model_para, 'MachLearnPara'))
                        error 'Model para not of type MachLearnPara'
                    end
                                        
                    % call calcETP
                    obj.print_calc_etp_seqs(al, stamp);
                    cmd = [FilesDirs.calc_etp_dir '/calc_ETP --tr --cons --al ' FilesDirs.calc_etp_al_filename(scenario) ' --stamp ' stamp ...
                           ' --home ' FilesDirs.home_dir()];  
                    if consts.use_scratch
                        cmd = [cmd ' --scratch'];
                    end                    
                    cmd = [cmd ' ' stamp];  
                    calc_etp_job = Job(consts.use_cluster, true, 'calcETP', stamp, cmd, []);
                    calc_etp_job.submit();

                    % carry out LDA
                    if (consts.use_cluster)
                        test_seq = obj;
                        save(FilesDirs.mach_learn_vars_filename(stamp), 'test_seq', 'model_para', 'al', 'consts', 'scenario');
                        ml_job = MachLearn.create_job(stamp);
                        ml_job.submit;
                        disp(manual_cluster);
                        cmd = [FilesDirs.wait_script_filename ' ' FilesDirs.ep_filename(stamp)]
                        wait_for_file_job = Job(true, true, 'waitFile', stamp, cmd, []);
                        wait_for_file_job.submit;
                    else
                        MachLearn.calc_em_probs(obj, model_para, al, stamp);    
                    end
                                       
                    % call jpHMM
                    if (consts.use_cluster)
                        jphmm_job = JPHMM.create_job(model_para, stamp, [calc_etp_job.cluster_id wait_for_file_job.cluster_id]);
                    else
                        jphmm_job = JPHMM.create_job(model_para, stamp, []);
                    end
                    jphmm_job.submit;
                    job = jphmm_job;
                otherwise
                    error 'Invalid strategy'
            end
        end
                
        % print seq to query seq file of jpHMM
        function print_jpHMM_query(obj, stamp)
            fid = FilesDirs.sec_fopen(FilesDirs.jpHMM_query_filename(stamp), 'w');
            name = 'Z.test';
            fprintf(fid, '>%s\n', name);
            fprintf(fid, '%s\n', obj.genome);
            fclose(fid);
        end        
    end
    
    methods (Access = private)
        % calculate transition freqs excluding seqs used by the test seq
        function calc_tr_freqs(obj, mod_sel_para, al, stamp)
            num_states = 2;
            fid = fopen(FilesDirs.tf_filename(stamp), 'w');
            if fid == -1 
                error(['Unable to open: ' FilesDirs.tf_filename(stamp)]) 
            end

            fprintf(fid, '%i\n', mod_sel_para.num_small_subs);
            fprintf(fid, '%i\n', mod_sel_para.num_large_subs);
            fprintf(fid, '%i\n', obj.len);
            
            subs = [mod_sel_para.small_subs mod_sel_para.large_subs];
            for i=1:length(subs)
                sub = char(subs{i});
                no_used_seqs = al.no_seqs_of_sub(sub);
                no_used_seqs = setdiff(no_used_seqs, obj.no_used_seqs);
                freqs = al.tr_freqs(obj.begin_pos, obj.end_pos, no_used_seqs);
                for j=1:obj.len-1
                    fprintf(fid, '%i %i', j-1, mod_sel_para.sub2c_idx(subs{i}));
                    for k=1:num_states
                        for l=1:num_states
                            fprintf(fid, ' %i', freqs(j, k, l));
                        end
                    end
                    fprintf(fid, '\n');
                end
            end          
        
            fclose(fid);
        end
        
        % print base data needed by calc_ETP
        function print_calc_etp_seqs(obj, al, stamp)
            global all_subs;
            global small_subs;
            global large_subs;
            fid = fopen(FilesDirs.calc_etp_seqs_filename(stamp), 'w');
            if fid == -1 
                error(['Unable to open: ' FilesDirs.calc_etp_seqs_filename(stamp)]) 
            end
            
            % print begin and end pos
            fprintf(fid, '%i\n', obj.begin_pos - 1);   
            fprintf(fid, '%i\n', obj.end_pos - 1);   
            
            % print small subs
            fprintf(fid, '>small\n');
%             for i=1:length(small_subs)
%                 fprintf(fid, '%s\n', small_subs{i});
%             end    
            for i=1:all_subs.num()
                fprintf(fid, '%s\n', all_subs.names{i});
            end    

            % print large subs
            fprintf(fid, '>large\n');     
%             for i=1:length(large_subs)
%                 fprintf(fid, '%s\n', large_subs{i});
%             end    
            % print numbers of sequences
            fprintf(fid, '>no_seqs\n');            
            curr_no_used_seqs = al.no_seqs();
            curr_no_used_seqs = setdiff(curr_no_used_seqs, obj.no_used_seqs);
            for i=1:length(curr_no_used_seqs)
                fprintf(fid, '%i\n', curr_no_used_seqs(i));   
            end            
            fclose(fid);
        end
    end
end


